(* Some examples of parametric polymorphism in OCaml. *)

(* Reverses a list. *)
let reverse xs0 =
  let rec loop acc xs =
    match xs with
    | []        -> acc
    | x :: xs'  -> loop (x :: acc) xs'
  in loop [] xs0
(* Inferred type scheme:
    reverse : 'a list -> 'a list *)

(* Sums a list of ints. *)
let sum xs0 =
  let rec loop acc = function
  | []      -> acc
  | x :: xs -> loop (x + acc) xs
  in loop 0 xs0
(* Inferred type scheme:
    sum : int list -> int *)

(* Sums a list of floats. *)
let sum_float xs0 =
  let rec loop acc = function
  | []      -> acc
  | x :: xs -> loop (x +. acc) xs
  in loop 0.0 xs0
(* Inferred type scheme:
    sum_float : float list -> float *)

(* Reduces a list, left associatively. *)
let foldl (plus, zero, xs0) =
  let rec loop acc = function
  | []      -> acc
  | x :: xs -> loop (plus (x, acc)) xs
  in loop zero xs0
(* Inferred type scheme:
    foldl : (('a * 'b -> 'b) * 'b * 'a list) -> 'b *)

(* Reduces a list, left associatively. *)
let foldl_curried plus zero xs0 =
  let rec loop acc = function
  | []      -> acc
  | x :: xs -> loop (plus x acc) xs
  in loop zero xs0
(* Inferred type scheme:
    foldl : ('a -> 'b -> 'b) -> 'b -> 'a list -> 'b *)

(* Sorts a list according to the given binary predicate. *)
let rec merge_sort (<) xs0 =

  let split =
    foldl_curried (fun x (xs1, xs2) -> (x :: xs2, xs1)) ([], []) in

  let rec merge xs1 xs2 = match xs1, xs2 with
    | x1 :: xs1', x2 :: xs2' ->
      if x1 < x2
        then x1 :: merge xs1' xs2
        else x2 :: merge xs1 xs2'
    | [], _ -> xs2
    | _, [] -> xs1 in

  let rec sort xs =
    match xs with
    | [] | [_] -> xs
    | _        ->
      let (xs1, xs2) = split xs in
        merge (sort xs1) (sort xs2) in

  sort xs0
(* Inferred type scheme:
    merge_sort : ('a -> 'a -> bool) -> 'a list -> 'a list *)
